/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <algorithm>
#include <vector>
#include <numeric>
#include <iostream>
#include <fstream>
#include <cmath>
#include <beattracker/beattracker.h>
//#include <boost/timer.hpp>

namespace beat_tracker {

using namespace std;

BeatTracker::BeatTracker(BeatTrackerInit bt_init)
#if EXTERNAL_FFT == 0
    :fft_(128)
#endif
{
    sampling_freq_ = bt_init.getSamplingFreq();
    fft_freq_      = bt_init.getWindowSize() / 2;
    shift_size_    = bt_init.getShiftSize();

    new_frames_    = bt_init.getWindowSize() / shift_size_;

    debug_flag_    = bt_init.getDebugFlag();

    log_prefix_    = bt_init.getLogPrefix() + "bt_";

		// メルフィルタバンク次元
		filter_num_    = 64;//bt_init.getFilterNum();
		//ずらす幅（スタート）
		//28->185bpmまで可
		cc_shift       = bt_init.getCcShiftStart();//44;
		//ずらす幅（おわり）
		cc_shift_end   = bt_init.getCcShiftEnd();  //87;
		//パターンマッチする幅
		cc_size        = cc_shift_end;//bt_init.getCcSize()*sampling_freq_/shift_size_;//cc_shift_end;
		std::cout<<"DEBUG"<<cc_size<<cc_shift<<cc_shift_end<<std::endl;
		//周波数（メル次元）スタート
		mel_f_start_   = 1;//bt_init.getMelFStart();//1;//40//54
		//周波数（メル次元）エンド
		mel_f_end_     = 63;//bt_init.getMelFEnd();//127;//63;//63
		//intervalの保存区間
		size_t interval_num = 13;//13;

    // frequency resolution
    const double freq_res = sampling_freq_ / (fft_freq_ * 2);

    // mel filter bank のそれぞれの周波数境界
    // bank vector<double> melfilterbank
    // const int filter_num
    double melf_max = 2595 * log10(1 + fft_freq_ * freq_res / 700);
    double mel_step_size = melf_max / filter_num_;
    melfilterbank_.resize(filter_num_ + 2);
    melfilterbank_[0] = 0;
    melfilterbank_[filter_num_ + 1] = fft_freq_;
    for (int i = 1; i < (filter_num_ + 1); ++i) {
        melfilterbank_[i] = (pow(10, (i * mel_step_size / 2595.0)) - 1) * 700
                            / freq_res;
        //cout << i << ":" << melfilterbank_[i] << endl;
    }//cout << melfilterbank_.size() << endl;

		//FreqBank
		SetFreqBank();

    beat_intervals.init(0.0);
    //ベクター準備
    vector<double> spec_init(fft_freq_, 0.0);
    spec_.init(spec_init);
    inc_spec_.init(spec_init);

    vector<double> mel_init(filter_num_, 0.0);
    mel_.init(mel_init);
    inc_mel_.init(mel_init);
    inc_mel_sum_.init(0.0);

    vector<double> close_beats_init(cc_shift_end, 0.0);
    close_beats_reliability_.init(close_beats_init);
    beats_reliability_mask_.init(close_beats_init);

    beats_reliability_.init(0.0);
    beats_reliability_sum_num_.init(0);

    music_reliability_.init(0.0);
    beat_intervals_temp.resize(interval_num, 0.0);
    music_reliability_temp.resize(interval_num, 0.0);

    // ディバック用ファイル
    ofs_spec.open                  ((log_prefix_+"spec.dat").c_str(),
                                    std::ios::out | std::ios::trunc);
    ofs_inc_spec.open              ((log_prefix_+"inc_spec.dat").c_str(),
                                    std::ios::out | std::ios::trunc);
    ofs_melfilterbank.open         ((log_prefix_+"melfilterbank.dat").c_str(),
                                    std::ios::out | std::ios::trunc);
    ofs_chroma_vector.open         ((log_prefix_+"chroma_vector.dat").c_str(),
                                    std::ios::out | std::ios::trunc);
		//ofs_mfcc.open                  ((log_prefix_+"mfcc.dat").c_str(), 
		//                                std::ios::out | std::ios::trunc);
		ofs_inc_mel.open               ((log_prefix_+"inc_mel.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
		ofs_interval_long_value.open   ((log_prefix_+"interval_long_value.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_interval_value.open        ((log_prefix_+"interval_value.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_interval_contrast.open     ((log_prefix_+"interval_contrast.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_interval_long_contrast.open((log_prefix_+"interval_long_contrast.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_beats_reliability_mask.open((log_prefix_+"beats_reliability_mask.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_beats_reliability.open     ((log_prefix_+"beats_reliability.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_beats_times.open           ((log_prefix_+"beats_times.dat").c_str(),
		                                std::ios::out | std::ios::trunc);

    //エコーキャンセル用ディバッカ
    ofs_flat_num.open              ((log_prefix_+"flat_num.dat").c_str(),
		                                std::ios::out | std::ios::trunc);
    ofs_music_reli.open            ((log_prefix_+"music_reli.dat").c_str(),
		                                std::ios::out | std::ios::trunc);


    /*
     * analyze用作業変数の初期化
     */
    t_current = new_frames_ / 2-1;
    beat_time = 0;
    beat_type = 0;
    beat_interval = 0;
    music_reliability = 0;
    music_signaloff_flag = false;

    cout << "beat_tracking initialize ended" << endl;
}

BeatTracker::~BeatTracker() {
}

BeatTrackerData BeatTracker::analyze(const std::vector<double>& new_spec)
{
    BeatTrackerData result;

    //	timer t;
//    boost::timer t;
    ++t_current;

    music_signaloff_flag = false;
    // spec update(time) スペクトルアップデート
    spec_.push_back(new_spec);

		//create chroma vector
		vector<double> chroma_tmp_ = MakeChromaVec(spec_);//chroma

    // mel filter bank(calc) メルフィルタバンク
    vector<double> melfilterbank_tmp_ = MelFilter(spec_);
    
    mel_.push_back(melfilterbank_tmp_);
    //-mel filter bank(calc) END

    // inc_mel calc メルスペクトル微分計算（ソーベルフィルタ）0,endは計算せず
    // filter
    int inc_filter[3][3] = { {-1, 0, 1}, {-2, 0, 2}, {-1, 0, 1}};
    vector<double> inc_mel_tmp_ = Filter3x3(mel_,inc_filter);
    inc_mel_.push_back(inc_mel_tmp_);
    // inc_mel calc END

    //相互相関スタート----------------------
    vector<double> interval_sum_value_ = STPattanMatching(inc_mel_);
    vector<double> interval_sum_value_pure_ = interval_sum_value_;

    //山の発見 Local maxima search
    //山の頂上のみ残す(並べ替えしやすいように) Leave only the local maxima (to make it easier to sort)
    vector<double> interval_value_top_(interval_sum_value_.size(), 0.0);
    if (interval_sum_value_[1] < interval_sum_value_[0])
        interval_value_top_[0] = interval_sum_value_[0];
    else
        interval_value_top_[0] = 0.0;

    if (*(interval_sum_value_.end() - 2) < *(interval_sum_value_.end() - 1))
        *(interval_value_top_.end() - 1) = *(interval_sum_value_.end() - 1);
    else
        *(interval_value_top_.end() - 1) = 0.0;

    for (size_t i = 1; i < interval_sum_value_.size() - 1; ++i) {
        if (interval_sum_value_[i - 1] < interval_sum_value_[i]
        && interval_sum_value_[i + 1] < interval_sum_value_[i])
            interval_value_top_[i] = interval_sum_value_[i];
        else
            interval_value_top_[i] = 0.0;
    }

    //離れている山を見つける Finding local maxima
    vector<double>::iterator l_top_tmp = max_element(interval_value_top_.begin(), interval_value_top_.end());

    /// by olivier
    double fMaxVal = -1;
    int i = 0;
    int nBestIndex=-1;
    for(vector<double>::iterator iter = interval_value_top_.begin(); iter < interval_value_top_.end(); iter++ )
    {
        if( fMaxVal < *iter )
        {
            nBestIndex = i;
            fMaxVal = *iter;
        }
        ++i;
    }
    ///


    double first_interval = l_top_tmp - interval_value_top_.begin();
    double first_interval_value = *l_top_tmp;
    *l_top_tmp = 0.0;

    double pre_interval = first_interval;

    double second_interval = -cc_shift;
    double second_interval_value = 0.0;

    static double old_first_interval_value = 0.0;
    static double old_pre_interval = 0;

    double flat_num = 0.0;

    if (first_interval_value == 0) {
        //全体が0の時 When it's zero
        pre_interval = old_pre_interval;
    } else {
        double y1, y2, y3, x2;
        x2 = 2;
        if (first_interval <= 2) {
            y1 = interval_sum_value_[first_interval + 2];
            y2 = interval_sum_value_[first_interval + 4];
            y3 = interval_sum_value_[first_interval];
        } else if (first_interval >= interval_sum_value_.size() - 3) {
            y1 = interval_sum_value_[first_interval - 2];
            y2 = interval_sum_value_[first_interval];
            y3 = interval_sum_value_[first_interval - 4];
        } else {
            y1 = interval_sum_value_[first_interval];
            y2 = interval_sum_value_[first_interval + 2];
            y3 = interval_sum_value_[first_interval - 2];
        }
        flat_num = (2 * y1 - y2 - y3) / (-2.0);

        //二番目の山 Second local maxima
        bool flag = true;
        while (flag)
        {
            l_top_tmp = max_element(interval_value_top_.begin(), interval_value_top_.end());

            //山がきちんととれていない（１番目に近い）場合はやり直し
            // Redo when the local-maxima was not accurately detected
            if (abs(l_top_tmp - interval_value_top_.begin() - first_interval) < 20 && *l_top_tmp != 0)
            {
                //loop again
                flag = true;
                *l_top_tmp = 0.0;
            }

            //山が小さい場合は終わり
            // Exit loop if local-maxima is small
            else if (*l_top_tmp < first_interval_value * 0.7)
            {
                //exit loop
                flag = false;
            }

            else
            {
                second_interval = l_top_tmp - interval_value_top_.begin();
                second_interval_value = *l_top_tmp;
                double diff = abs(first_interval - second_interval);

                if (diff> 19)
                {
                    for (; diff < (cc_shift - 5);) diff *= 2;

                    double diff2 = diff * 2;
                    diff -= cc_shift;
                    diff2 -= cc_shift;
                    //diff = 可能性の高いインターバル Interval of high likelihood

                    if (abs(diff - second_interval) < abs(diff - first_interval) && abs(diff - second_interval) < 5)
                    {
                        pre_interval = second_interval;
                    }
                    else if(abs(diff2 - second_interval) < abs(diff2 - first_interval) && abs(diff2	- second_interval) < 5)
                    {
                        pre_interval = second_interval;
                    }
                    else if (abs(diff - first_interval) >= 5 && abs(diff2 - first_interval) >= 5)
                    {
                        //firstもsecondも離れている場合,３番を試す
                        // If both the first and the second local-maxima are distant, try the third one
                        diff = abs(first_interval - second_interval);
                        for (; diff < (cc_shift - 5);) diff *= 3;

                        diff -= cc_shift;

                        if (abs(diff - second_interval) < abs(diff - first_interval) && abs(diff - second_interval) < 5)
                        {
                            pre_interval = second_interval;
                        }
                    }
                }

                flag = false;
            }

        }


        //連続処理 Continuous processing
        //昔の信頼度を残す Decrease previous confidence
        old_first_interval_value *= 0.85;

        //現在の信頼度と比べて Compare with the current confidence
        if (old_first_interval_value > first_interval_value)
        {
            //現在の方が低い場合、昔のまま使う
            // If the current confidence is low, use the previous one
            pre_interval = old_pre_interval;
            first_interval_value = old_first_interval_value;
        }
        else
        {
            double diff = abs(old_pre_interval - pre_interval);

            //昔との差が２倍３倍になっているとき
            // When the difference with the previous interval is two or three times

            if (diff> 19)
            {
                for (; diff < (cc_shift - 5);) diff *= 2;

                double diff2 = diff*2 - cc_shift;
                diff -= cc_shift;
                if (abs(diff - old_pre_interval) < 5)
                {
                    if(abs(diff2 - pre_interval) >= 5)
                    {

                        //昔の方がメインである可能性が高いため、昔の方を使う
                        pre_interval = diff;//old_pre_interval;
                        first_interval_value = old_first_interval_value;
                    }
                }
                else if(abs(diff - pre_interval) >= 5)
                {
                    //そうでないとき３倍を試す
                    //					cout << "new diff   : " << abs(old_pre_interval
                    //							- pre_interval) << endl;
                    //					cout << "new old_pre: " << old_pre_interval << endl;
                    //					cout << "new pre    : " << pre_interval << endl;
                    //					cout << "new abs    : " << abs(diff - old_pre_interval)
                    //							<< endl;

                    diff = abs(old_pre_interval - pre_interval);

                    for (; diff < (cc_shift - 5);) diff *= 3;

                    diff -= cc_shift;
                    //３のdiffをとる
                    if (abs(diff - old_pre_interval) < 5)
                    {
                        //昔の方がメインである可能性が高い
                        pre_interval = diff;//old_pre_interval;
                        first_interval_value = old_first_interval_value;
                    }
                }
            }
        }
    }

    //全体が０の時 When the total is zero
    vector<double>::iterator l_pure_max = max_element(interval_sum_value_pure_.begin(), interval_sum_value_pure_.end());
    if (*l_pure_max == 0)
    {
        pre_interval = -5;
        music_signaloff_flag = true;
    }

    old_pre_interval = pre_interval;
    old_first_interval_value = first_interval_value;
    //シフト（周期方向）終わり

    //相互相関終わり----------------------
    // interval update アップデート
    beat_intervals.push_back(pre_interval + cc_shift);
    // interval update アップデート END

    static double music_reliability_range = 200.0;

    size_t flat_number = 0;
    for (size_t i = 1; i < static_cast<size_t>(music_reliability_range+1); ++i) {
        if (beat_intervals[0] + 5> beat_intervals[i] && beat_intervals[i] >	beat_intervals[0] - 5)
            flat_number++;
    }
    if (music_signaloff_flag) {
        music_reliability_.push_back(0.0);
    } else {
        music_reliability_.push_back(flat_number / music_reliability_range);
    }
		if(debug_flag_)
			ofs_flat_num << flat_number << endl;
    //インターバル処理終了

    // mel合計（ビート時刻検出用）アップデート
    inc_mel_sum_.push_back(accumulate((inc_mel_[0].begin() + mel_f_start_ - 1),
    (inc_mel_[0].begin() + mel_f_end_ - 1), 0.0));
    // mel合計（ビート時刻検出用）アップデート END


    //ビート時刻検出スタート
    // onset→近接ビート信頼度
    vector<double> close_beats_reliability_tmp(cc_shift_end, 0.0);
    double interval_tmp = beat_intervals[0];
    //切り上げ
    int interval_ceil = ceil(interval_tmp);
    //切り下げ
    int interval_floor = floor(interval_tmp);

    for (int i = 0; i < interval_tmp; ++i)
    {
			//beat 列
			double sumsum_ 
				= inc_mel_sum_[i] 
				+ inc_mel_sum_[i + interval_floor] * (interval_ceil - interval_tmp) 
				+ inc_mel_sum_[i + interval_ceil]  * (interval_tmp  - interval_floor);
			close_beats_reliability_tmp[i] = sumsum_;
    }
    for (int i = interval_tmp; i < cc_shift_end; ++i)
    {
			close_beats_reliability_tmp[i] = 0;
    }
    close_beats_reliability_.push_back(close_beats_reliability_tmp);

		if (debug_flag_)
		{
			for (int i = 0; i < cc_shift_end; ++i)
			{				
				ofs_inc_spec << close_beats_reliability_tmp[i] << " ";
			}
			ofs_inc_spec << endl;
    }
	
    //近接ビート信頼度→信頼度マスク
    vector<double> beats_reliability_mask_tmp(cc_shift_end, 0.0);

    //参照するまえのビートの数
    int ref_beats_num=3;
    //持っていない時間のビートは計算しない
    if(beat_interval!=0)
    {
        int ref_num = (t_current - new_frames_ / 2)/beat_interval;
        if(ref_num<3)
					ref_beats_num=ref_num;
    }
    else
    {
        ref_beats_num = 0;
    }

    double target_time       = beat_intervals[0];
    int    target_time_ceil  = ceil(target_time);
    int    target_time_floor = floor(target_time);
    double next_interval_tmp = 0.0;

    for (int j = 0; j < ref_beats_num; ++j) {
			for (int i = 0; i < cc_shift_end; ++i) {
				if (target_time == target_time_ceil)
					beats_reliability_mask_tmp[i]
						+= close_beats_reliability_[target_time][i];
				else
					beats_reliability_mask_tmp[i]
						+= close_beats_reliability_[target_time_floor][i]
						* (target_time_ceil - target_time)
						+ close_beats_reliability_[target_time_ceil][i]
						* (target_time - target_time_floor);
			}
			if ((target_time_ceil - target_time)
					< (target_time - target_time_floor)) {
				next_interval_tmp = beat_intervals[target_time_ceil];
			} else {
				next_interval_tmp = beat_intervals[target_time_floor];
			}
			target_time       += next_interval_tmp;
			target_time_ceil   = ceil (target_time);
			target_time_floor  = floor(target_time);
    }
		if (debug_flag_)
		{
			for (size_t i = 0; i < beats_reliability_mask_tmp.size(); ++i)
        ofs_beats_reliability_mask 
					<< beats_reliability_mask_tmp[i]
					* close_beats_reliability_tmp[i] << " ";
			ofs_beats_reliability_mask << endl;
		}
    beats_reliability_mask_.push_back(beats_reliability_mask_tmp);

    //Σ(近接ビート信頼度*信頼度マスク)→ビート時刻尤度
    beats_reliability_.push_back(0.0);
    beats_reliability_sum_num_.push_back(0);
    //		vector<double>::reverse_iterator br_rit = beats_reliability_.rbegin();
    //		for (size_t i=0; i<beats_reliability_mask_tmp.size(); ++i) {
    //			*br_rit += (*br_rit*beats_reliability_sum_num_[i]
    //					+beats_reliability_mask_tmp[i]
    //							*close_beats_reliability_tmp[i])
    //					/(beats_reliability_sum_num_[i]+1.0);
    //			beats_reliability_sum_num_[i] += 1;
    //			++br_rit;
    //		}
    for (size_t i = 0; i < beats_reliability_mask_tmp.size(); ++i) {
        beats_reliability_[i] 
					+= (beats_reliability_[i] * beats_reliability_sum_num_[i] 
							+ 
							beats_reliability_mask_tmp[i] * close_beats_reliability_tmp[i])
					//	+ close_beats_reliability_tmp[i])
					/ (beats_reliability_sum_num_[i] + 1.0);
        beats_reliability_sum_num_[i] += 1;
    }

    //山の頂上のみ残す
    //		vector<double>::reverse_iterator detect_top_tmp =
    //				beats_reliability_.rbegin()+cc_shift_end;
    //		if (*detect_top_tmp<*(detect_top_tmp+1))
    //			*detect_top_tmp=0.0;
    //		if (*detect_top_tmp>*(detect_top_tmp+1))
    //			*(detect_top_tmp+1)=0.0;
    if (beats_reliability_[cc_shift_end] < beats_reliability_[cc_shift_end + 1])
        beats_reliability_[cc_shift_end] = 0.0;
    else
        beats_reliability_[cc_shift_end + 1] = 0.0;

    //		if (debug_flag_) // originally comment out
    //			ofs_beats_reliability
    //					<< beats_reliability_[beats_reliability_.size()-cc_shift_end-2] << endl;
		if (debug_flag_)
			ofs_beats_reliability << beats_reliability_[cc_shift_end + 2] << endl;
    //ビート時刻確定

    double interval_ave = beat_interval;

    //ビートタイム捜査
    size_t temp_num = ceil(interval_ave);
    vector<double> beats_reliability_temp_(temp_num, 0.0);
    for (size_t i = 0; i < beats_reliability_temp_.size(); ++i)
        beats_reliability_temp_[i] = beats_reliability_[cc_shift_end + 1 + i];

    //データがしっかりそろったら
    if (t_current> (cc_shift_end + cc_size) && ref_beats_num==3) {
        static bool first_detect = false;
        if (beat_time + interval_ave * 3 / 2 < t_current - cc_shift_end) {
            //一番高い山を探す
            vector<double>::iterator detect_top_start =
            beats_reliability_temp_.begin();
            vector<double>::iterator detect_top_end =
            beats_reliability_temp_.end();
            //				cout << "first mount : ";
            //				for(vector<double>::reverse_iterator it = detect_top_start;it<detect_top_end;++it)
            //					cout << *it << " ";
            //				cout << endl;

            vector<double>::iterator beat_reli_max = max_element(
                detect_top_start, detect_top_end);
            double beat_reli_first = *beat_reli_max;

            //一番高い山が０でない場合
            if (beat_reli_first != 0)
            {
                *beat_reli_max = 0.0;
                //高い山の周りを０に
                size_t flat_num = floor(interval_ave / 3);
                for (vector<double>::iterator it = beat_reli_max - flat_num; it
                < beats_reliability_temp_.end() && it < it + flat_num; it++) {
                    if (it - beats_reliability_temp_.begin() >= 0)
                        *it = 0.0;
                }
                //１の近くにいないか
                vector<double>::iterator beat_reli_max_second = max_element(
                    detect_top_start, detect_top_end);
                while (*beat_reli_max_second != 0 && abs(beat_reli_max
                - beat_reli_max_second) < interval_ave / 3) {
                    *beat_reli_max_second = 0.0;
                    cout << "cal max in time sec" << endl;
                    beat_reli_max_second = max_element(detect_top_start,
                    detect_top_end);
                }
                double beat_reli_second = *beat_reli_max_second;
                *beat_reli_max_second = 0.0;
                //山の周りを０に
                for (vector<double>::iterator it = beat_reli_max_second
                - flat_num; it < beats_reliability_temp_.end() && it
                < it + flat_num; it++) {
                    if (it - beats_reliability_temp_.begin() >= 0)
                        *it = 0.0;
                }
                //3番目に高い山
                vector<double>::iterator beat_reli_max_third = max_element(
                    detect_top_start, detect_top_end);
                //１の近くにいないか
                while (*beat_reli_max_third != 0 && (abs(beat_reli_max
                - beat_reli_max_third) < interval_ave / 3 || abs(
                    beat_reli_max_second - beat_reli_max_third)
                < interval_ave / 3)) {
                    *beat_reli_max_third = 0.0;
                    cout << "cal max in time third" << endl;
                    beat_reli_max_third = max_element(detect_top_start,
                    detect_top_end);
                }
                *beat_reli_max_second = beat_reli_second;
                *beat_reli_max = beat_reli_first;
                //高い山を二つ探し終える

                //確定範囲を指定
                double beat_range_start = t_current - cc_shift_end
                - interval_ave * 2 / 3.0;
                double beat_range_end = t_current - cc_shift_end - interval_ave
                / 3.0;

                //第1候補
                double beat_time1 = t_current - (beat_reli_max
                - beats_reliability_temp_.begin()) - cc_shift_end;
                //第2候補
                double beat_time2 = t_current - (beat_reli_max_second
                - beats_reliability_temp_.begin()) - cc_shift_end;
                //第3候補
                double beat_time3 = t_current - (beat_reli_max_third
                - beats_reliability_temp_.begin()) - cc_shift_end;
                //どれでもないとき
                double beat_time_etc = t_current - cc_shift_end - interval_ave
                / 2.0;

                bool beat1_in_flag = false;
                bool beat2_in_flag = false;
                bool beat3_in_flag = false;
                //候補間の比
                double order = *beat_reli_max / *beat_reli_max_second;
                double order_third = *beat_reli_max / *beat_reli_max_third;

                static double old_interval_ave = 0;
                //インターバルの差が大きいとき
                if (abs(old_interval_ave - interval_ave)> 5) {
                    //true:初めてではない　false:初めて
                    first_detect = false;
                }
                //
                old_interval_ave = interval_ave;
                static size_t not_detect_count = 0;
//					if (not_detect_count> 3) {
//						first_detect = false;
//						not_detect_count = 0;
//					}
                //				if(beat_time1 > beat_range_start && beat_time1
                //										< beat_range_end && *beat_reli_max!=0){
                //					beat1_in_flag = true;
                //				}
                //				if (beat_time2 > beat_range_start && beat_time2
                //										< beat_range_end && first_detect && *beat_reli_max_second!=0 && order < 100.0){
                //					 beat2_in_flag = true;
                //				}
                //				if (beat_time3 > beat_range_start && beat_time3
                //										< beat_range_end && first_detect && *beat_reli_max_third!=0 && order_third < 100.0){
                //					beat3_in_flag = true;
                //				}
                //
                if (*beat_reli_max != 0) {
                    beat1_in_flag = true;
                }
                if (first_detect && *beat_reli_max_second != 0 && order < 50.0) {
                    beat2_in_flag = true;
                }
                if (first_detect && *beat_reli_max_third != 0 && order_third
                < 50.0) {
                    beat3_in_flag = true;
                }

                //				cout << "first : " <<  beat_time1
                //				<< " " << *beat_reli_max
                //				<< "\t second : " <<  beat_time2
                //				<< " " << *beat_reli_max_second
                //				<< "\t third : " <<  beat_time3
                //				<< " " << *beat_reli_max_third << endl;
                //平均との差を取る
                double diff_time1 = abs(beat_time1 - beat_time_etc);
                double diff_time2 = abs(beat_time2 - beat_time_etc);
                double diff_time3 = abs(beat_time3 - beat_time_etc);

                //差の一番小さい時間を選ぶ
                static int detect_count=0;

                if (beat1_in_flag && beat2_in_flag && beat3_in_flag) {
                    ++detect_count;
                    if (diff_time1 <= diff_time2 && diff_time1 <= diff_time3) {
                        beat_time = beat_time1;
                    } else if (diff_time2 <= diff_time1 && diff_time2
                    <= diff_time3) {
                        beat_time = beat_time2;
                    } else {
                        beat_time = beat_time3;
                    }

                } else if (beat1_in_flag && beat2_in_flag) {
                    ++detect_count;
                    if (diff_time1 <= diff_time2) {
                        beat_time = beat_time1;
                    } else {
                        beat_time = beat_time2;
                    }

                } else if (beat1_in_flag) {
//						detect_count=0;
                    ++detect_count;
                    beat_time = beat_time1;

                } else {
                    cout << "beat not detect" << endl;
                    not_detect_count++;
                    beat_time = beat_time_etc;
                }

                if ((beat_time < beat_range_start || beat_time> beat_range_end) && first_detect)
                {
                    cout << "beat not detect out of range : ";
                    if(not_detect_count < min(detect_count/5.0,1.0)) {
                        not_detect_count++;
                        beat_time = beat_time_etc;
                        cout << "remain" << endl;
                    } else {
                        beat_time = beat_time1;
                        cout << "change : " << detect_count << ", " << not_detect_count << endl;
                        detect_count=0;
                        not_detect_count=0;
                    }
                } else {
                    not_detect_count=0;
                }
                first_detect = true;

                //if (debug_flag_)
								//	ofs_beats_times << beat_time << " " << 1 << endl;
            }
        }
    }

    //ビート時刻検出エンド

    //インターバルを考慮して、最新のビート時刻を出す。
    int next_beat_time = beat_time;
    bool next_beat_flag = true;
    if(beat_time != 0) {
        int count = 0;
        while(next_beat_flag) {
            int now_interval = beat_intervals[t_current-next_beat_time];
            int temp_beat_time = now_interval + next_beat_time;
            count++;
            if(count> 50) {
                cout << "temp_beat_time : " << temp_beat_time << endl;
                cout << "t_current : " << t_current << endl;
                cout << "now_interval : " << now_interval << endl;
                cout << "next_beat_time : " << next_beat_time << endl;
            }
            if(temp_beat_time < t_current && now_interval != 0) {
                int next_interval = beat_intervals[t_current-temp_beat_time];
                if(next_interval < now_interval && next_interval != 0) {
                    temp_beat_time = next_beat_time + next_interval;
                } else if(next_interval == now_interval) {
                    if((next_beat_time+cc_shift_end)< t_current) {
                        next_interval = beat_intervals[t_current-(next_beat_time+cc_shift_end)];
                        if(now_interval < next_interval) {
                            temp_beat_time = next_beat_time+next_interval;
                        }
                    }
                }
                if(count> 50) {
                    cout << "in temp_beat_time : " << temp_beat_time << endl;
                    cout << "in t_current : " << t_current << endl;
                    cout << "in now_interval : " << now_interval << endl;
                    cout << "in next_beat_time : " << next_beat_time << endl;
                }
                if(temp_beat_time < t_current) {
                    next_beat_time = temp_beat_time;
                } else {
                    next_beat_flag = false;
                }
            } else {
                next_beat_flag = false;
            }
        }
        //cout <<"end next time"<<endl;
    }
    // interval_value contrast
    // インターバルのコントラストを計算
    double contrast = 0.0;
    vector<double>::iterator l_max = max_element(interval_sum_value_.begin(),
    interval_sum_value_.end());
    vector<double>::iterator l_min = min_element(interval_sum_value_.begin(),
    interval_sum_value_.end());
    if ((*l_max + *l_min) == 0)
        contrast = 0.0;
    else
        contrast = (*l_max - *l_min) / (*l_max + *l_min);
    //contrast = l_max-l_min;
    //		double pre_interval = cc_shift + (l_max
    //		- interval_sum_value_.begin()) + (*(l_max-1)
    //		-*(l_max+1)) /(2* *(l_max-1) - 4* *l_max+2* *(l_max+1));
		
    static size_t count_a = 0;
    count_a++;
    if (debug_flag_)
			ofs_interval_long_contrast //*(l_max - 1) + *(l_max + 1) + 2 * *l_max // orignal c-o
				<< t_current*shift_size_/sampling_freq_ << " " 
				<< pre_interval    + cc_shift << " " 
				<< first_interval  + cc_shift << " " 
				<< second_interval + cc_shift << " "
				<< first_interval_value  << " " 
				<< second_interval_value << endl;
    //				cout << "output_time : " << t_current*512.0/44100.0
    //				<< " | " << count_a << " / " << t_current << endl;

    // interval並べ替え
    for (size_t i = 0; i < beat_intervals_temp.size(); ++i) {
        beat_intervals_temp[i] = beat_intervals[i];
    }
    sort(beat_intervals_temp.begin(), beat_intervals_temp.end());
    beat_interval = beat_intervals[0];//beat_intervals_temp[beat_intervals_temp.size()/2];

    // 音楽信頼度
    for (size_t i = 0; i < music_reliability_temp.size(); ++i) {
        music_reliability_temp[i] = music_reliability_[i];
    }
    //double music_reliability = 0.0;
    sort(music_reliability_temp.begin(), music_reliability_temp.end());
    music_reliability = music_reliability_[0];
//		music_reliability = music_reliability_temp[music_reliability_temp.size() / 2];

		if (debug_flag_)
			ofs_music_reli 
				<< t_current * shift_size_ / sampling_freq_ << " " 
				<< music_reliability << " " 
				<< beat_interval << endl;

    //	cout << "time : " << t_current*shift_size_/sampling_freq_
    //			<< "\t interval : " << beat_interval << endl;
    // beat_time(relative),beat_type,beat_interval
    //	cout << "anain  end : " << t.elapsed() << endl;

    //test
    //	beat_time = 1;
    //	beat_interval = 55;
    //BeatTrackerData result; olivier
    result.setInterval(beat_interval);
    result.setBeatTimes(beat_time, t_current);
    result.setRecentBeatTime(next_beat_time,t_current);
    result.setBeatType(beat_type);
    result.setMusicReliability(music_reliability);

		result.setSTPMVector(interval_sum_value_);
		result.setOnsetVector(inc_mel_tmp_);
		result.setChromaVector(chroma_tmp_); //chroma
    //			return make_tuple(t_current-beat_time, beat_time/*beat_type*/,
    //					beat_interval);
    return result;
}
//________________________analyze_______________________\\


void BeatTracker::SetFreqBank()
{
	static const double pi = 3.1415926;
	vec_f_cent_bank_.reserve(fft_freq_);
	vec_f_cent_h_.reserve(fft_freq_);
	vec_f_cent_c_.reserve(fft_freq_);
	//vec_f_cent_bank_.reserve(fft_freq_);
	double const_value = 440 * pow(2, 3 / 12.0 - 5);
	for (size_t f = 0; f < fft_freq_; ++f) {
		//cent に変換
		double f_cent  = 1200 * log2(f / const_value);
		size_t notenum = (int)(f_cent / 100.0 + 0.5) + 5;
		int    h       = notenum / 12;//if(h < 1 || 9 < h)continue;
		int    c       = notenum % 12; 
		double f_c_h   = 1200 * h + 100 * c;
		double bpf_ch  = 0.50 - (0.50 * cos(2.0 * pi * (f_cent - (f_c_h - 100)) / 200.0));
		vec_f_cent_bank_.push_back(f_cent);
		vec_f_cent_h_   .push_back(h);
		vec_f_cent_c_   .push_back(c);
	}

}

std::vector<double> BeatTracker::MakeIncSpec(RingIBuffer<std::vector<double> >& spec) {
    // inc_spec
    vector<double> inc_spec_tmp_(spec[0].size(), 0.0);
    for (size_t f=0; f < spec[0].size(); f++) {
        double inc_spec_tmp = spec[0][f]-spec[1][f];
        if(inc_spec_tmp>0)
            inc_spec_tmp_[f]=inc_spec_tmp;
        else
            inc_spec_tmp_[f]=0;
    }
    return inc_spec_tmp_;
    //inc_spec_.push_back(inc_spec_tmp_);
}

std::vector<double> BeatTracker::MakeChromaVec(RingIBuffer<std::vector<double> >& spec) {
    static const double pi = 3.1415926;
    // 12 dim chroma vector 12次元クロマベクトル（test）3-8 * 12
    vector<double> chroma_vector_tmp_(12, 0.0);
    double tmp_chroma = 0;
		//cout << spec[0].size() << endl; // sepc bug
    for (size_t f = 0; f < spec[0].size(); f++) {
			//cent に変換
			double f_cent = vec_f_cent_bank_[f];
#if 1
		// window : rectangle
			size_t notenum = (int)(f_cent / 100.0 + 0.5) + 5;
			int h = vec_f_cent_h_[f];
			int c = vec_f_cent_c_[f];
			if(h < 1 || 9 < h)continue;
			chroma_vector_tmp_[c] += spec[0][f];
#else
		// window : Hamming
				/*
				for (size_t h = 3; h <= 8; ++h) {
					for (size_t c = 0; c < 12; ++c) {
               double f_c_h = 1200 * h + 100 * c;
                if (f_c_h - 100 < f_cent && f_cent < f_c_h + 100) {
                    double bpf_ch = 0.50 - (0.50 * cos(2.0 * pi * (f_cent - (f_c_h - 100)) / 200.0));
                    chroma_vector_tmp_[ c      ] +=      bpf_ch  * spec[0][f];
										chroma_vector_tmp_[(c+1)%12] += (1 - bpf_ch) * spec[0][f];
										cout << h << "," << c << ":" << (int)(f_cent/100)/12 << "," << ((int)(f_cent/100))%12 << " ";
										cout << bpf_ch << ",";  // sepc bug
										goto OUT;
                }
            }
        }
				*/
        size_t notenum = (int)(f_cent / 100.0 + 0.5) + 5;
        int h = notenum / 12;
				int c = notenum % 12;
				double f_c_h = 1200 * h + 100 * c;
				if(h < 1 || 9 < h)continue;
				//cout << f_cent << "," << h << "," << c <<endl;
				
				double bpf_ch = 0.50 - (0.50 * cos(2.0 * pi * (f_cent - (f_c_h - 100)) / 200.0));
				chroma_vector_tmp_[ c      ] +=      bpf_ch  * spec[0][f];
				chroma_vector_tmp_[(c+1)%12] += (1 - bpf_ch) * spec[0][f];
#endif
    }

#if 0
    for (size_t i = 0; i < 12 ; i++)
    	tmp_chroma += SQAURE(chroma_vector_tmp_[i]);
#else
		for (size_t i = 0; i < 12 ; i++)
		{
			if(tmp_chroma < chroma_vector_tmp_[i])
				tmp_chroma  = chroma_vector_tmp_[i];
		}
#endif

    for (size_t i = 0; i < 12 ; i++)
    	chroma_vector_tmp_[i] /= tmp_chroma;
    
		if (debug_flag_)
		{
			for (size_t i = 0; i < chroma_vector_tmp_.size(); ++i) 
			{
				ofs_chroma_vector << chroma_vector_tmp_[i] << " ";
			}
			ofs_chroma_vector << endl;
		}
    return chroma_vector_tmp_;
}
std::vector<double> BeatTracker::MelFilter(RingIBuffer<std::vector<double> >& spec) {
    // mel filter bank(calc) メルフィルタバンク
    vector<double> melfiltered_spec(melfilterbank_.size()-2, 0.0);
    //for (int i = 1; i < (filter_num + 1); ++i) {
    for (size_t i= 1; i<(melfilterbank_.size()-1); ++i) 
    {
        for (size_t f = 0; f < spec_[0].size(); f++) 
        {
            if (melfilterbank_[i - 1] <= f && f <= melfilterbank_[i]) 
            {
                melfiltered_spec[i - 1] += spec_[0][f] 
                                         * (f                 - melfilterbank_[i - 1]) 
                                         / (melfilterbank_[i] - melfilterbank_[i - 1]);
            } 
            else if (melfilterbank_[i] < f && f <= melfilterbank_[i + 1]) 
            {
                melfiltered_spec[i - 1] += spec_[0][f]
                                         * (melfilterbank_[i + 1] - f) 
                                         / (melfilterbank_[i + 1] - melfilterbank_[i]);
            }
        }
    }
    if (debug_flag_)
		{
			for (size_t i= 1; i<(melfilterbank_.size()-1); ++i) 
			{
					ofs_melfilterbank << melfiltered_spec[i - 1] << " ";
					//cout << i - 1 << ":" << melfiltered_spec[i - 1] << " ";
			}
			ofs_melfilterbank << endl;
			//cout << melfilterbank_.size()-1 << ":" << melfiltered_spec[melfilterbank_.size()-1] << endl;
		}

    return melfiltered_spec;
}
std::vector<double> BeatTracker::Filter3x3(RingIBuffer<std::vector<double> >& input_signal,int filter[][3]) {
    // inc_mel calc メルスペクトル微分計算（ソーベルフィルタ）0,endは計算せず
    size_t filter_num = input_signal[0].size();
    vector<double> output_signal(filter_num, 0.0);
    for (size_t i = 1; i < filter_num - 1; ++i) {
        //			for (int i=mel_f_start; i<mel_f_end; ++i) {
        output_signal[i] = filter[0][0] * input_signal[2][i - 1]
                         + filter[0][1] * input_signal[1][i - 1]
                         + filter[0][2] * input_signal[0][i - 1]            
                         + filter[1][0] * input_signal[2][i] 
                         + filter[1][1] * input_signal[1][i]        
                         + filter[1][2] * input_signal[0][i] 
                         + filter[2][0] * input_signal[2][i + 1] 
                         + filter[2][1] * input_signal[1][i + 1] 
                         + filter[2][2] * input_signal[0][i + 1];
        if (output_signal[i] < 0)
            output_signal[i] = 0.0;
    }
		if (debug_flag_)
		{
			for (size_t i = 1; i < filter_num - 1; ++i) {
        ofs_inc_mel << output_signal[i] << " ";
				//cout << i << ":" << output_signal[i] << " ";
			}
			ofs_inc_mel << endl;
    //cout << endl;
		}

    return output_signal;
}
std::vector<double> BeatTracker::STPattanMatching(RingIBuffer<std::vector<double> >& input_signal) {
    //相互相関スタート----------------------
    vector<double> output_signal(cc_shift_end - cc_shift, 0.0);
		//シフト（周期方向）
    for (int shift_ft = cc_shift; shift_ft < cc_shift_end; ++shift_ft) {
        double tmp_value0 = 0.0;
        double tmp_value1 = 0.0;
        double tmp_value2 = 0.0;

        // f 周波数方向
        for (int j = mel_f_start_; j < mel_f_end_; ++j) {
            // time 時間方向
            for (int i = 0; i < cc_size; ++i) {

                //NCC^2
                tmp_value0 += input_signal[i][j] * input_signal[i + shift_ft][j];
                tmp_value1 += input_signal[i][j] * input_signal[i][j];
                tmp_value2 += input_signal[i + shift_ft][j]
                * input_signal[i + shift_ft][j];

            }// time end 時間方向終わり
        }// f end 周波数方向おわり

        //NCC^2
        if (tmp_value1 == 0 || tmp_value2 == 0)
            output_signal[shift_ft - cc_shift] = 0.0;
        else
            output_signal[shift_ft - cc_shift] = tmp_value0 * tmp_value0
            / tmp_value1 / tmp_value2;
    }

    if (debug_flag_)
		{
			for (int shift_ft = cc_shift; shift_ft < cc_shift_end; ++shift_ft) {
					ofs_interval_long_value << output_signal[shift_ft - cc_shift] << " ";
			}
			ofs_interval_long_value << endl;
		}
		return output_signal;
}
/*void BeatTrackerData::sendmessage_ioi(udp::socket& socket, double ioi, double next)
{
	string first_message("a," + lexical_cast<string>(next) + "," + lexical_cast<string>(ioi) );
	cout << first_message << " " << first_message.size() << endl;
	socket.send(buffer(first_message, first_message.size()));//error std::exception 12 03 26
	return;
}

void BeatTrackerData::sendmessage_start(udp::socket& socket)
{
	string start("s");
	socket.send(buffer(start, start.size()));
	return;
}

void BeatTrackerData::sendmessage_end(udp::socket& socket)
{
  string end("e");
	socket.send(buffer(end, end.size()));
	return;
}*/
}
